Object subclass: #ExampleServer
	instanceVariableNames: 'stdio '
	classVariableNames: 'ClientProcesses ServerProcess '
	poolDictionaries: ''
	category: 'ESUG-UnixServer'!
ExampleServer comment:
'Expression evaluation server for Unix.  See the ESUG summer school ''95 proceedings for explanation.

This class, and all its methods, are Copyright (c) 1995 Ian Piumarta (ian.piumarta@inria.fr).  Use of this class, and all its methods, is hereby granted for any purpose provided that this paragraph appears in any software, and its supporting documentation, that is in any way based on, or otherwise derived from, this class or its methods.
'!


!ExampleServer methodsFor: 'initialize-release'!

setStdio: anExternalStream
	stdio := anExternalStream! !

!ExampleServer methodsFor: 'accessing'!

stdio
	^stdio! !

!ExampleServer methodsFor: 'services-capitalisation'!

capitalise
	stdio nextPutAll: stdio contents asUppercase!

capitalise: aStream
	aStream nextPutAll: aStream contents asUppercase! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ExampleServer class
	instanceVariableNames: ''!


!ExampleServer class methodsFor: 'class initialization'!

initialize
	"Initialize the server. If the class has been previously initialized, we may 
	need to kill an existing server and its clients."

	ServerProcess notNil ifTrue: [self killServer].
	ClientProcesses isNil
		ifTrue: [ClientProcesses := IdentitySet new]
		ifFalse: [self killClientProcesses].
	(ObjectMemory dependents includes: self)
		ifFalse: [ObjectMemory addDependent: self]! !

!ExampleServer class methodsFor: 'instance creation'!

newClientStdio: anExternalStream
	^super new setStdio: anExternalStream! !

!ExampleServer class methodsFor: 'server process control'!

killServer
	"Kill the server process."

	ServerProcess isNil ifTrue: [self error: 'there is no server process to kill'].
	ServerProcess terminate.
	self notifyServer: 'killed'.
	ServerProcess := nil!

resumeServer
	"Resume the server process."

	ServerProcess isNil ifTrue: [self error: 'there is no server process to resume'].
	ServerProcess resume.
	self notifyServer: 'resumed'.!

startServer
	"Start a new server process."

	ServerProcess isNil ifFalse: [self error: 'a server process already exists'].
	ServerProcess := [self serverLoop] newProcess.
	ServerProcess priority: Processor userInterruptPriority.
	ServerProcess resume.
	self notifyServer: 'started'.!

suspendServer
	"Suspend the server process."

	ServerProcess isNil ifTrue: [self error: 'there is no server process to suspend'].
	ServerProcess suspend.
	self notifyServer: 'suspended'.! !

!ExampleServer class methodsFor: 'client process control'!

killClientProcesses
	"Terminate all the client processes."

	ClientProcesses do: [ :process |
		process terminate.
		self notifyClient: 'killed' for: process.
		ClientProcesses remove: process]! !

!ExampleServer class methodsFor: 'server loop'!

dispatchClientOn: clientSocket

	| stdio command client process |
	stdio := clientSocket readAppendStream.
	command := stdio through: Character cr.
	client := self newClientStdio: stdio.
	ClientProcesses add: (process :=
		[
			[	self evaluateSafely: command for: client.
				stdio flush; close.
			] valueNowOrOnUnwindDo:
				[ClientProcesses remove: process.
				self notifyClient: 'exited' for: process.
				clientSocket close]
		] newProcess).
	self notifyClient: 'started' for: process.
	process priority: Processor userSchedulingPriority.
	process resume.!

serverLoop
	"Create a socket and listen for connections. When a connection 
	arrives, create a stream connected to the client's stdin/out, read
	the command to be avaluated, and create a client process
	connected to the stdio stream to #perform: the command."

	| serverSocket |
	serverSocket := UnixSocketAccessor
				family: (SocketAddress domainCodeFromName: #afInet)
				type: UnixSocketAccessor sockStream.
	[	serverSocket
			setOptionsLevel: UnixSocketAccessor solSocket
				name: UnixSocketAccessor soReuseaddr
				value: 1;
			bindTo: (IPSocketAddress
						hostAddress: IPSocketAddress thisHost
						port: self serverPort)
	] valueOnUnwindDo:
		[serverSocket close.
		self error: 'failed to create server socket'].
	
	[	serverSocket listenFor: 1.
		[true] whileTrue: [self dispatchClientOn: serverSocket accept]
	] valueNowOrOnUnwindDo: [serverSocket close]! !

!ExampleServer class methodsFor: 'configuration constants'!

serverPort
	"Return the port number on which the server listens for connections"

	^1966! !

!ExampleServer class methodsFor: 'snapshot'!

returnFromSnapshot
	ServerProcess isNil ifFalse: [self killServer; killClientProcesses; startServer]!

update: aspect with: parameter from: sender 
	(sender == ObjectMemory and: [aspect == #returnFromSnapshot])
		ifTrue: [self returnFromSnapshot]! !

!ExampleServer class methodsFor: 'private'!

evaluateSafely: command for: client
	| unique |
	unique := 'unique'.
	^Smalltalk errorSignal
		handle: [ :ex || message |
			message := ex originator == unique
				ifTrue: ['*** syntax error']
				ifFalse: ['*** ' ,  ex errorString , Character cr asSymbol ,
						'*** for receiver: ' , ex originator printString].
			client stdio nextPutAll: message; cr; flush.
			ex returnWith: '']
		do: [Compiler new evaluate: command in: nil to: client notifying: unique ifFail: nil]!

notifyClient: action for: client

	Transcript cr; show: 'Client process ' , client asOop printString , ' [' , client printString , '] ' , action; endEntry!

notifyServer: action

	Transcript cr; show: 'Server process ' , ServerProcess asOop printString , ' [' , ServerProcess printString , '] ' , action; endEntry! !

ExampleServer subclass: #ExampleServerForResult
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'ESUG-UnixServer'!
ExampleServerForResult comment:
'Expression evaluation server for Unix.  See the ESUG summer school ''95 proceedings for explanation.

This class, and all its methods, are Copyright (c) 1995 Ian Piumarta (ian.piumarta@inria.fr).  Use of this class, and all its methods, is hereby granted for any purpose provided that this paragraph appears in any software, and its supporting documentation, that is in any way based on, or otherwise derived from, this class or its methods.
'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ExampleServerForResult class
	instanceVariableNames: ''!


!ExampleServerForResult class methodsFor: 'server loop'!

dispatchClientOn: clientSocket

	| stdio command client process |
	stdio := clientSocket readAppendStream.
	command := stdio through: Character cr.
	client := self newClientStdio: stdio.
	ClientProcesses add: (process :=
		[
			[	| result |
				result := (self evaluateSafely: command for: client).
				result = '' ifFalse: [stdio nextPutAll: result printString; cr].
			] valueNowOrOnUnwindDo:
				[ClientProcesses remove: process.
				self notifyClient: 'exited' for: process.
				stdio flush; close.
				clientSocket close]
		] newProcess).
	self notifyClient: 'started' for: process.
	process priority: Processor userSchedulingPriority.
	process resume.! !
ExampleServer initialize!

