/* stclient.c
 *
 * Connects to a local Smalltalk process on a well-known port, sends
 * the command-line arguments to Smalltalk as a string terminated with
 * a newline, and then loops copying stdin to the input side os the
 * Smalltalk socket and the  output side of the socket to stdout.  The
 * process exits when the socket is closed down from the Smalltalk side.
 *
 * ANSI-philes beware: this code is cheerfully traditional.
 */

#include <stdio.h>
#include <sysexits.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <netdb.h>
#include <netinet/in.h>

#define SMALLTALK_HOST  "localhost"
#define SMALLTALK_PORT  1966

#define BUFFER_SIZE     4096

#define ERR(X) { perror(X); exit(EX_UNAVAILABLE); }
#define FAIL(X) { fprintf(stderr, "%s failed\n", (X)); exit(EX_UNAVAILABLE); }

/* open a TCP connection to the given host and port number
 */
static tcp_connect(name, port)
char *name;
{
  struct sockaddr_in addr;
  struct hostent *hp;
  int stream, true= 1;
  int retry= 10;

  while (1) { /* loop until connection established */
    /* setup the foreign socket address */
    addr.sin_family= AF_INET;
    addr.sin_port= htons(port);
    if (!(hp= gethostbyname(name))) ERR(name);
    bcopy(hp->h_addr, &addr.sin_addr, sizeof(struct in_addr));
    /* create a reusable socket for a TCP connection */
    if ((stream= socket(AF_INET, SOCK_STREAM, 0)) < 0) ERR("socket");
    if (setsockopt(stream, SOL_SOCKET, SO_REUSEADDR, &true, sizeof(true)) < 0)
      ERR("setsockopt");
    /* try to connect, returning connected socket if successful */
    if (connect(stream, &addr, sizeof(struct sockaddr_in)) >= 0)
      return stream; /* succeed */
    /* for end may be busy, so don't give up immediately if refused */
    if (errno == ECONNREFUSED && retry--) sleep(1);
    else {
      close(stream);
      return -1; /* fail */
    }
  }
}

/* send strings from "strvec" (terminated with a null pointer) to
 * "destfd", in a single message, separated by spaces, and terminated
 * by a newline.
 */
static sendstrings(destfd, strvec)
char **strvec;
{
  int len;
  char *obuf, *pchar, **pstr;

  /* allocate a buffer of sufficient size for strings */
  for (len= 0, pstr= strvec; *pstr; pstr++) len+= strlen(*pstr)+1;
  if (!(pchar= obuf= (char *)malloc(len))) FAIL("malloc");
  /* linearise the strings into the buffer, and terminate with newline */
  for (pstr= strvec; *pstr; pstr++) {
    pchar+= strlen(strcpy(pchar, *pstr));
    *pchar++= ' ';
  }
  *(pchar-1)= '\n';
  /* atomically write buffer to destination, being pessimistic about failure */
  if (len != write(destfd, obuf, len)) FAIL("atomic write of string vector");
  if(!free(obuf)) FAIL("free");
}

/* copy a chunk of data from "fromfd" to "tofd", returning -1 if the
 * "fromfd" has reached EOF.
 */
static ship(fromfd, tofd)
{
  static char iobuf[BUFFER_SIZE];
  int nread, nwritten;
  char *optr= iobuf;
  if ((nread= read(fromfd, iobuf, BUFFER_SIZE)) < 0) ERR("read");
  if (nread == 0) return -1;  /* fail: connection reset by peer */
  /* loop, copying chunks of data, until input drained */
  while (nread > 0) {
    if ((nwritten= write(tofd, optr, nread)) < 0) ERR("write");
    if (nwritten) nread-= nwritten, optr+= nwritten;
    else {
      fprintf(stderr, "write returned zero -- why?\n");
      sleep(1);
    }
  }
  return 0;  /* succeed */
}

/* asynchronously connect a unidirectional "source" and "sink" to a
 * bidirectional socket, shutting down the input side of the socket
 * when "source" is exhausted, and returning when the socket is closed
 * down by the foreign host.
 */
static to_from_duplex(source, sink, sock)
{
  fd_set openfds;

  /* we're interested in input events on either source or socket */
  FD_ZERO(&openfds);
  FD_SET(source, &openfds);
  FD_SET(sock, &openfds);
  /* drain the socket to the sink, until the socket is disconnected */
  while (FD_ISSET(sock, &openfds)) {
    fd_set readfds;
    int nready;

    /* wait for data on either of the inputs, ignoring interrupted syscalls */
    readfds= openfds;
    while ((nready= select(NOFILE, &readfds, 0, 0)) < 0 && errno == EINTR);
    if (nready < 0) ERR("select");
    /* ship the available data in the appropriate direction(s) */
    if (FD_ISSET(0, &readfds) && ship(source, sock)) { /* EOF on source */
      FD_CLR(0, &openfds);
      shutdown(sock, 1); /* inform peer that input side is shut down */
    }
    if (FD_ISSET(sock, &readfds) && ship(sock, sink)) { /* connection reset */
      FD_CLR(sock, &openfds);
      shutdown(sock, 2); /* shut down both sides of socket */
    }
  }
}

/* here we go...
 */
main(argc, argv)
char **argv;
{
  int sock;

  /* try to connect to the Smalltalk execution server */
  if ((sock= tcp_connect(SMALLTALK_HOST, SMALLTALK_PORT)) < 0) {
    fprintf(stderr, "failed to connect to host: %s\n", SMALLTALK_HOST);
    exit(EX_UNAVAILABLE);
  }
  /* send the command line arguments (i.e. the command to be evaluated) */
  argv[argc]= 0;
  sendstrings(sock, argv+1);
  /* connect stdin and stdout to the Smalltalk process */
  to_from_duplex(fileno(stdin), fileno(stdout), sock);
  /* communication finished: close socket and terminate successfully */
  if (close(sock)) ERR("close");
  return EX_OK;
}
